int tty_class_init(void);
int tty_init(void);

int main () {
  /* Declaration of local variables */
  struct file my_file;
  struct inode my_inode;
  struct task_struct my_current;
  struct sighand_struct my_sighand;
  struct signal_struct my_signal;

  /* Initialize the helper structures that are used in the
     initialization phase of the driver itself. This is a very
     simplified version of the real kernel data structures.
   */
  my_signal.leader=0;
  my_signal.tty = NULL;
  my_current.sighand = & my_sighand;
  my_current.signal = &my_signal;
  current = &my_current;
  my_file.f_u.fu_list.next = &my_file;
  my_file.f_u.fu_list.prev = &my_file;
  my_file.f_path.mnt = 0;
  my_file.f_path.dentry = 0;
  my_file.f_op = &def_chr_fops;
  my_file.f_count = 1;
  my_file.f_version = 0;
  my_file.private_data = 0;
  my_file.f_flags = 0;
  my_file.f_mode = 0;

  /* initialization of the modules */
  chrdev_init();
  console_init();
  tty_class_init();
  tty_init();
  leonuart_init();

  unsigned int status = 0;

  /* Compute an initial status for the hardware. This will be inlined
     during the final analysis to separate states between the various
     possible hardware status.
   */
  uart_random_status();

  /* attach the inode to the driver */
  my_inode.i_rdev = leon_reg.tty_driver->cdev.dev;
  my_inode.i_cdev = &(leon_reg.tty_driver->cdev);

  /* open the driver */
  my_file.f_op->open(&my_inode,&my_file);
  leon_reg.tty_driver->cdev.ops->open(&my_inode,&my_file);

  /* sample buffer for multi-characters operations */
  const char* my_buf = "My Buffer";
  int length_buf = 9;

  struct tty_struct* tty = leon_ports[0].port.info->tty;

  /* maximum number of steps in practice. Changes between initial phase (we
     need to visit only once each instruction) and final analysis (the scenario
     has more than one step).
  */
  int exit_loop = max_step();

  // register initial state
  calmos_hook (-1);

  /* put some non-determinism in mctrl, otherwise constant propagation
     gets to aggressive.
   */
  leon_uart_put_ctrl(&leon_ports[0].port,Frama_C_nondet(0,1));

  /* the loop itself */
  while (exit_loop--) {
    int choice = Frama_C_interval(0,15);
    switch (choice) {
    case 0: {
      /* for tiocmset, the argument is important. We make several choices */
      int tmp1 = Frama_C_nondet(-1,0);
      /* the disjunction in the assertion forces the value analysis to keep
         states separated.
      */
      /*@ assert tmp1 == -1 || tmp1 == 0; */
      int tmp2 = Frama_C_nondet(-1,0);
      /*@ assert tmp2 == -1 || tmp2 == 0; */
      leon_reg.tty_driver->tiocmset(tty,&my_file,tmp1,tmp2);
      break;
    }
#ifndef RESTRICT
      /* this case, as well as others, are not considered when -restrict
         is on in script.ml
      */
    case 1: {
      /*@ assert leon_reg.state->info->xmit.head == 0 || leon_reg.state->info->xmit.head > 0; */
      leon_reg.tty_driver->write(tty,my_buf,length_buf);
      break;
    }
#endif
    case 2: {
      /* we can not send the NULL character, as it is a marker
         that no character is to be sent... */
      leon_reg.tty_driver->put_char(tty,Frama_C_interval(1,127));
      break;
    }
#ifndef RESTRICT
    case 3: {
      leon_reg.tty_driver->flush_chars(tty);
      break;
    }
    case 4: {
      leon_reg.tty_driver->write_room(tty);
      break;
    }
    case 5: {
      leon_reg.tty_driver->chars_in_buffer(tty);
      break;
    }
    case 6: {
      leon_reg.tty_driver->throttle(tty);
      break;
    }
    case 7: {
      leon_reg.tty_driver->unthrottle(tty);
      break;
    }
    case 8: {
      leon_reg.tty_driver->flush_buffer(tty);
      break;
    }
#endif
    case 9: {
      leon_reg.tty_driver->send_xchar(tty,Frama_C_interval(1,127));
      break;
    }
#ifndef RESTRICT
    case 10: {
      leon_reg.tty_driver->wait_until_sent
        (tty,Frama_C_interval(0,INT_MAX));
      break;
    }
    case 11: {
      leon_reg.tty_driver->start(tty);
      break;
    }
    case 12: {
      leon_reg.tty_driver->stop(tty);
      break;
    }
    case 13: {
      leon_reg.tty_driver->hangup(tty);
      break;
    }
#endif
    case 14: {
      /* Simulation of the hardware interrupt indicating that some
         chars need to be read on the UART
      */
      leon2uart_int(leon_ports[0].port.irq,&leon_ports[0].port,0);
      break;
    }
    default: {
      exit_loop = 0;
      break;
    }
    }
    calmos_hook(choice);
    uart_random_status();
  }
  return 0;
}
