/*
  stubs replacing unneeded kernel definitions
 */

#ifndef __STUBS
#define __STUBS

#include "stdio.h"
#include "stdlib.h"
#include "string.h"
#include "__fc_builtin.h"
#include "__fc_define_size_t.h"
#include "sys/types.h"

#include "termbits.h"

#ifndef NULL
#define NULL 0
#endif

#define WORK_DATA_INIT() { .counter = 0 }

#define PREPARE_WORK(_work, _func)                              \
        do {                                                    \
                (_work)->func = (_func);                        \
        } while (0)

#define INIT_WORK(_work, _func)                                         \
        do {                                                            \
                (_work)->data = (atomic_long_t) WORK_DATA_INIT();       \
                INIT_LIST_HEAD(&(_work)->entry);                        \
                PREPARE_WORK((_work), (_func));                         \
        } while (0)


#define __bitwise__
#if 0
// [VP] Disabled: we simulate status changes in the UART by hand.
#define __iomem volatile
#endif
#define __iomem
#define __force
#define __init
#define __exit
#define __user
#define KERN_INFO
#define KERN_NOTICE
#define KERN_WARNING
#define KERN_ERR
#define KERN_ALERT
#define KERN_DEBUG

#define WARN_ON(x)
#define WARN_ON_ONCE(x)

#define PAGE_SIZE (512)
#define IRQ_HANDLED (1)

#define _NSIG           64
#define _NSIG_BPW      32
#define _NSIG_WORDS     (_NSIG / _NSIG_BPW)

#define O_NONBLOCK 000040000
#define O_NOCTTY        00000400

#define ERESTARTSYS (512)
#define ENOIOCTLCMD (515)

#define SEND_SIG_NOINFO ((struct siginfo *) 0)
#define SEND_SIG_PRIV   ((struct siginfo *) 1)
#define SEND_SIG_FORCED ((struct siginfo *) 2)

#define SIGHUP           1
#define SIGINT           2
#define SIGQUIT          3
#define SIGILL           4
#define SIGTRAP          5
#define SIGABRT          6
#define SIGIOT           6
#define SIGBUS           7
#define SIGFPE           8
#define SIGKILL          9
#define SIGUSR1         10
#define SIGSEGV         11
#define SIGUSR2         12
#define SIGPIPE         13
#define SIGALRM         14
#define SIGTERM         15
#define SIGSTKFLT       16
#define SIGCHLD         17
#define SIGCONT         18
#define SIGSTOP         19
#define SIGTSTP         20
#define SIGTTIN         21
#define SIGTTOU         22
#define SIGURG          23
#define SIGXCPU         24
#define SIGXFSZ         25
#define SIGVTALRM       26
#define SIGPROF         27
#define SIGWINCH        28
#define SIGIO           29
#define SIGPOLL         SIGIO
/*
#define SIGLOST         29
*/
#define SIGPWR          30
#define SIGSYS          31
#define SIGUNUSED       31

/* These should not be considered constants from userland.  */
#define SIGRTMIN        32
#define SIGRTMAX        _NSIG

#define SIG_IGN ((__force __sighandler_t)1)

#define NR_IRQS (32)
#define SERIAL_IO_MEM (2)
#define ASYNC_BOOT_AUTOCONF 0x10000000
#define THIS_MODULE "leon"
#define TTY_MAJOR 4
#define TTYAUX_MAJOR            5

#define TASK_RUNNING 0
#define TASK_INTERRUPTIBLE 1

#define TCGETS          0x5401
#define TCSETS          0x5402 /* Clashes with SNDCTL_TMR_START sound ioctl */
#define TCSETSW         0x5403
#define TCSETSF         0x5404
#define TCGETA          0x5405
#define TCSETA          0x5406
#define TCSETAW         0x5407
#define TCSETAF         0x5408
#define TCSBRK          0x5409
#define TCXONC          0x540A
#define TCFLSH          0x540B
#define TIOCEXCL        0x540C
#define TIOCNXCL        0x540D
#define TIOCSCTTY       0x540E
#define TIOCGPGRP       0x540F
#define TIOCSPGRP       0x5410
#define TIOCOUTQ        0x5411
#define TIOCSTI         0x5412
#define TIOCGWINSZ      0x5413
#define TIOCSWINSZ      0x5414
#define TIOCMGET        0x5415
#define TIOCMBIS        0x5416
#define TIOCMBIC        0x5417
#define TIOCMSET        0x5418
#define TIOCGSOFTCAR    0x5419
#define TIOCSSOFTCAR    0x541A
#define FIONREAD        0x541B
#define TIOCINQ         FIONREAD
#define TIOCLINUX       0x541C
#define TIOCCONS        0x541D
#define TIOCGSERIAL     0x541E
#define TIOCSSERIAL     0x541F
#define TIOCPKT         0x5420
#define FIONBIO         0x5421
#define TIOCNOTTY       0x5422
#define TIOCSETD        0x5423
#define TIOCGETD        0x5424
#define TCSBRKP         0x5425  /* Needed for POSIX tcsendbreak() */
/* #define TIOCTTYGSTRUCT 0x5426 - Former debugging-only ioctl */
#define TIOCSBRK        0x5427  /* BSD compatibility */
#define TIOCCBRK        0x5428  /* BSD compatibility */
#define TIOCGSID        0x5429  /* Return the session ID of FD */
#define TCGETS2         _IOR('T', 0x2A, struct termios2)
#define TCSETS2         _IOW('T', 0x2B, struct termios2)
#define TCSETSW2        _IOW('T', 0x2C, struct termios2)
#define TCSETSF2        _IOW('T', 0x2D, struct termios2)
#define TIOCGRS485      0x542E
#define TIOCSRS485      0x542F
#define TIOCGPTN        _IOR('T', 0x30, unsigned int)
                                /* Get Pty Number (of pty-mux device) */
#define TIOCSPTLCK      _IOW('T', 0x31, int)  /* Lock/unlock Pty */
#define TCGETX          0x5432 /* SYS5 TCGETX compatibility */
#define TCSETX          0x5433
#define TCSETXF         0x5434
#define TCSETXW         0x5435

#define FIONCLEX        0x5450
#define FIOCLEX         0x5451
#define FIOASYNC        0x5452
#define TIOCSERCONFIG   0x5453
#define TIOCSERGWILD    0x5454
#define TIOCSERSWILD    0x5455
#define TIOCGLCKTRMIOS  0x5456
#define TIOCSLCKTRMIOS  0x5457
#define TIOCSERGSTRUCT  0x5458 /* For debugging only */
#define TIOCSERGETLSR   0x5459 /* Get line status register */
#define TIOCSERGETMULTI 0x545A /* Get multiport config  */
#define TIOCSERSETMULTI 0x545B /* Set multiport config */

#define TIOCMIWAIT      0x545C  /* wait for a change on serial input line(s) */
#define TIOCGICOUNT     0x545D  /* read serial port inline interrupt counts */
#define TIOCGHAYESESP   0x545E  /* Get Hayes ESP configuration */
#define TIOCSHAYESESP   0x545F  /* Set Hayes ESP configuration */
#define FIOQSIZE        0x5460

/* Used for packet mode */
#define TIOCPKT_DATA             0
#define TIOCPKT_FLUSHREAD        1
#define TIOCPKT_FLUSHWRITE       2
#define TIOCPKT_STOP             4
#define TIOCPKT_START            8
#define TIOCPKT_NOSTOP          16
#define TIOCPKT_DOSTOP          32

#define TIOCSER_TEMT    0x01    /* Transmitter physically empty */


#define GFP_KERNEL 0xD0u
#define GFP_ATOMIC 0x20u

#define CAP_SYS_ADMIN (21)

#define INIT_C_CC "\003\034\177\025\004\0\1\0\021\023\032\0\022\017\027\026\0"

#define ASYNC_CLOSING_WAIT_NONE 65535

#define TTY_DRIVER_INSTALLED            0x0001
#define TTY_DRIVER_RESET_TERMIOS        0x0002
#define TTY_DRIVER_REAL_RAW             0x0004
#define TTY_DRIVER_DYNAMIC_DEV          0x0008
#define TTY_DRIVER_DEVPTS_MEM           0x0010
#define TTY_DRIVER_HARDWARE_BREAK       0x0020

/* tty driver types */
#define TTY_DRIVER_TYPE_SYSTEM          0x0001
#define TTY_DRIVER_TYPE_CONSOLE         0x0002
#define TTY_DRIVER_TYPE_SERIAL          0x0003
#define TTY_DRIVER_TYPE_PTY             0x0004
#define TTY_DRIVER_TYPE_SCC             0x0005  /* scc driver */
#define TTY_DRIVER_TYPE_SYSCONS         0x0006

/* system subtypes (magic, used by tty_io.c) */
#define SYSTEM_TYPE_TTY                 0x0001
#define SYSTEM_TYPE_CONSOLE             0x0002
#define SYSTEM_TYPE_SYSCONS             0x0003
#define SYSTEM_TYPE_SYSPTMX             0x0004

/* pty subtypes (magic, used by tty_io.c) */
#define PTY_TYPE_MASTER                 0x0001
#define PTY_TYPE_SLAVE                  0x0002

/* serial subtype definitions */
#define SERIAL_TYPE_NORMAL      1

#define TTY_DRIVER_MAGIC                0x5402

#define TTY_THROTTLED 		0	/* Call unthrottle() at threshold min */
#define TTY_IO_ERROR 		1	/* Canse an I/O error (may be no ldisc too) */
#define TTY_OTHER_CLOSED 	2	/* Other side (if any) has closed */
#define TTY_EXCLUSIVE 		3	/* Exclusive open mode */
#define TTY_DEBUG 		4	/* Debugging */
#define TTY_DO_WRITE_WAKEUP 	5	/* Call write_wakeup after queuing new */
#define TTY_PUSH 		6	/* n_tty private */
#define TTY_CLOSING 		7	/* ->close() in progress */
#define TTY_LDISC 		9	/* Line discipline attached */
#define TTY_HW_COOK_OUT 	14	/* Hardware can do output cooking */
#define TTY_HW_COOK_IN 		15	/* Hardware can do input cooking */
#define TTY_PTY_LOCK 		16	/* pty private */
#define TTY_NO_WRITE_SPLIT 	17	/* Preserve write boundaries to driver */
#define TTY_HUPPED 		18	/* Post driver->hangup() */
#define TTY_FLUSHING		19	/* Flushing to ldisc in progress */
#define TTY_FLUSHPENDING	20	/* Queued buffer flush pending */

#define TTY_WRITE_FLUSH(tty) tty_write_flush((tty))

extern void tty_write_flush(struct tty_struct *);

#define POLLIN          0x0001
#define POLLPRI         0x0002
#define POLLOUT         0x0004
#define POLLERR         0x0008
#define POLLHUP         0x0010
#define POLLNVAL        0x0020
#define POLLWRNORM      0x0100
#define POLLRDNORM      0x0040

#define DEFINE_MUTEX(x) struct mutex x = { .lock = 0 }
#define DEFINE_SPINLOCK(x) spinlock_t x
#define LIST_HEAD(x) struct list_head x = { &(x), &(x) }
#define DECLARE_WAIT_QUEUE_HEAD(x) LIST_HEAD(x)
#define DECLARE_WAITQUEUE(x,y) struct list_head x;

#define CIRC_SPACE_TO_END(head,tail,size) \
        ({int end = (size) - 1 - (head); \
          int n = (end + (tail)) & ((size)-1); \
          n <= end ? n : end+1;})

#define CON_ENABLED (4)

#define MINORBITS (20)
#define MKDEV(ma,mi) (((ma) << MINORBITS) | (mi))
#define MINORMASK       ((1U << MINORBITS) - 1)
#define MAJOR(dev)      ((unsigned int) ((dev) >> MINORBITS))
#define MINOR(dev)      ((unsigned int) ((dev) & MINORMASK))

#define KOBJ_NAME_LEN 20
#define BUS_ID_SIZE KOBJ_NAME_LEN

#define BUS_NOTIFY_DEL_DEVICE           0x00000002 /* device removed */
#define BUS_NOTIFY_BOUND_DRIVER         0x00000003 /* driver bound to device */
#define BUS_NOTIFY_UNBIND_DRIVER        0x00000004 /* driver to be unbound*/


#define S_IRUGO         (S_IRUSR|S_IRGRP|S_IROTH)

#define HZ 100
#define LDISC_FLAG_DEFINED      0x00000001
#define printk(...)
#define sprintf(...)
#define pr_debug(...)
#define snprintf(...)
#define __stringify_1(x)        #x
#define __stringify(x)          __stringify_1(x)

#define set_current_state(x)

#define do_each_pid_task(pid, type, task)                               \
        do {                                                            \
                struct hlist_node *pos___;                              \
                if ((pid) != NULL)                                      \
                        hlist_for_each_entry_rcu((task), pos___,        \
                                &(pid)->tasks[type], pids[type].node) {

                        /*
                         * Both old and new leaders may be attached to
                         * the same pid in the middle of de_thread().
                         */
#define while_each_pid_task(pid, type, task)                            \
                                if (type == PIDTYPE_PID)                \
                                        break;                          \
                        }                                               \
        } while (0)

#define do_each_thread(g, t) \
        for (g = t = &init_task ; (g = t = next_task(g)) != &init_task ; ) do

#define while_each_thread(g, t) \
        while ((t = next_thread(t)) != g)

#define hlist_for_each_entry_rcu(tpos, pos, head, member)                \
        for (pos = rcu_dereference((head)->first);                       \
                pos && ({ prefetch(pos->next); 1; }) &&                  \
                ({ tpos = hlist_entry(pos, typeof(*tpos), member); 1; }); \
                pos = rcu_dereference(pos->next))

#define rcu_dereference(p) p
#define prefetch(x) x

#define next_task(p) list_entry(rcu_dereference((p)->tasks.next), struct task_struct, tasks)

#define min(x,y) ({ \
         typeof(x) _x = (x);     \
         typeof(y) _y = (y);     \
         (void) (&_x == &_y);            \
         _x < _y ? _x : _y; })

enum pid_type
{
  PIDTYPE_PID,
  PIDTYPE_PGID,
  PIDTYPE_SID,
  PIDTYPE_MAX
};

enum kobject_action {
        KOBJ_ADD,
        KOBJ_REMOVE,
        KOBJ_CHANGE,
        KOBJ_MOVE,
        KOBJ_ONLINE,
        KOBJ_OFFLINE,
        KOBJ_MAX
};

int kobject_add(struct kobject *kobj, struct kobject *parent, const char* fmt);

struct list_head {
  struct list_head* next, *prev;
};

typedef unsigned int __u32;
typedef unsigned long u64;
typedef unsigned __bitwise__ gfp_t;

typedef int spinlock_t;
struct __wait_queue_head {
        spinlock_t lock;
        struct list_head task_list;
};
typedef struct __wait_queue_head wait_queue_head_t;
typedef int dev_t;
typedef int irqreturn_t;
typedef int loff_t;

typedef int (*initcall_t)(void);
typedef struct { volatile int counter; } atomic_t;

typedef struct { long counter; } atomic_long_t;

struct completion {
        unsigned int done;
        wait_queue_head_t wait;
};

struct kref {
        atomic_t refcount;
};

void kref_init(struct kref *);

extern void init_waitqueue_head(wait_queue_head_t *q);

static inline void init_completion(struct completion *x)
{
        x->done = 0;
        init_waitqueue_head(&x->wait);
}

struct hlist_node {
        struct hlist_node *next, **pprev;
};

struct klist_node;
struct klist {
        spinlock_t              k_lock;
        struct list_head        k_list;
        void                    (*get)(struct klist_node *);
        void                    (*put)(struct klist_node *);
};


extern void klist_init(struct klist * k, void (*get)(struct klist_node *),
                       void (*put)(struct klist_node *));

struct klist_node {
        struct klist            * n_klist;
        struct list_head        n_node;
        struct kref             n_ref;
        struct completion       n_removed;
};

extern void klist_add_tail(struct klist_node * n, struct klist * k);
extern void klist_add_head(struct klist_node * n, struct klist * k);

extern void klist_del(struct klist_node * n);
extern void klist_remove(struct klist_node * n);

extern int klist_node_attached(struct klist_node * n);



struct pid_link
{
        struct hlist_node node;
        struct pid *pid;
};

// the original declaration would be very difficult for frama-c to interpret
// correctly.
//extern initcall_t __con_initcall_start[],__con_initcall_end[];
extern initcall_t __con_initcall_start[], *__con_initcall_end;

#define __NFDBITS       (8 * sizeof(unsigned long))

#define __FD_SETSIZE    1024

#define __FDSET_LONGS   (__FD_SETSIZE/__NFDBITS)

typedef struct {
        unsigned long fds_bits [__FDSET_LONGS];
} fd_set;

struct embedded_fd_set {
        unsigned long fds_bits[1];
};

struct rcu_head {
        struct rcu_head *next;
        void (*func)(struct rcu_head *head);
};

struct fdtable {
        unsigned int max_fds;
        struct file ** fd;      /* current fd array */
        fd_set *close_on_exec;
        fd_set *open_fds;
        struct rcu_head rcu;
        struct fdtable *next;
};

#define BITS_PER_LONG 32

#define NR_OPEN_DEFAULT BITS_PER_LONG

struct files_struct {
  /*
   * read mostly part
   */
        atomic_t count;
        struct fdtable *fdt;
        struct fdtable fdtab;
  /*
   * written part on a separate cache line in SMP
   */
        spinlock_t file_lock;
        int next_fd;
        struct embedded_fd_set close_on_exec_init;
        struct embedded_fd_set open_fds_init;
        struct file * fd_array[NR_OPEN_DEFAULT];
};

#define files_fdtable(files) (rcu_dereference((files)->fdt))

typedef struct {
        unsigned long sig[_NSIG_WORDS];
} sigset_t;

//NdV: far from complete
struct task_struct {
  volatile long state;    /* -1 unrunnable, 0 runnable, >0 stopped */
  void *stack;
  atomic_t usage;
  unsigned int flags;     /* per process flags, defined below */
  unsigned int ptrace;
  int lock_depth;
  struct signal_struct* signal;
  struct sighand_struct *sighand;
  struct list_head thread_group;
  struct pid_link pids[PIDTYPE_MAX];
  struct files_struct *files;
  struct list_head tasks;
  sigset_t blocked, real_blocked;
};

extern struct task_struct init_task;
//NdV: Frama-C does not really what to do with this volatile...
/*volatile*/ struct task_struct* current;

struct poll_table_struct;

struct module;

static inline int try_module_get(struct module *module)
{
        return 1 ;
}
static inline void module_put(struct module *module)
{
}
static inline void __module_get(struct module *module)
{
}

static inline void __list_add(struct list_head *new,
                              struct list_head *prev,
                              struct list_head *next)
{
        next->prev = new;
        new->next = next;
        new->prev = prev;
        prev->next = new;
}

static inline void __list_del(struct list_head * prev, struct list_head * next)
{
        next->prev = prev;
        prev->next = next;
}

static inline void list_add(struct list_head *new, struct list_head *head)
{
        __list_add(new, head, head->next);
}

static inline int list_empty(const struct list_head *head)
{
        return head->next == head;
}

static inline void INIT_LIST_HEAD(struct list_head *list)
{
        list->next = list;
        list->prev = list;
}

static inline void list_del_init(struct list_head *entry)
{
        __list_del(entry->prev, entry->next);
        INIT_LIST_HEAD(entry);
}

static inline void list_add_tail(struct list_head *new, struct list_head *head)
{
        __list_add(new, head->prev, head);
}


struct serial_icounter_struct {
  int cts, dsr, rng, dcd;
  int rx, tx;
  int frame, overrun, parity, brk;
  int buf_overrun;
  int reserved[9];
};

struct kobject {
	const char		*name;
	struct kref		kref;
	struct list_head	entry;
  	struct kobject		*parent;
  	struct kset		*kset;
        struct kobj_type	*ktype;
  //	struct sysfs_dirent	*sd;
  unsigned int state_initialized/*:1*/;
  unsigned int state_in_sysfs/*:1*/;
  unsigned int state_add_uevent_sent/*:1*/;
  unsigned int state_remove_uevent_sent/*:1*/;
};

struct kset {
        struct list_head list;
        spinlock_t list_lock;
        struct kobject kobj;
  //struct kset_uevent_ops *uevent_ops;
};

void kset_init(struct kset *k);

struct kobj_type {
	void (*release)(struct kobject *kobj);
  //struct sysfs_ops *sysfs_ops;
	struct attribute **default_attrs;
};

struct sysfs_ops {
	ssize_t	(*show)(struct kobject *, struct attribute *,char *);
	ssize_t	(*store)(struct kobject *,struct attribute *,const char *, size_t);
};

#define device_init_wakeup(dev,val)
#define wake_up_interruptible(queue)
#define free_irq(irq,port)

static inline void set_dev_node(struct device *dev, int node) { }

typedef unsigned short mode_t;

struct attribute {
	const char		*name;
	struct module		*owner;
	mode_t			mode;
};

struct cdev {
  struct kobject kobj;
  struct module *owner;
  const struct file_operations *ops;
  struct list_head list;
  dev_t dev;
  unsigned int count;
};

struct work_struct;
typedef void (*work_func_t)(struct work_struct *work);

struct work_struct {
  atomic_long_t data;
  struct list_head entry;
  work_func_t func;
};

struct timer_list {
  struct list_head entry;
  long expires;
  long data;
};

struct delayed_work {
  struct work_struct work;
  struct timer_list timer;
};

typedef long off_t;

typedef off_t loff_t;

typedef long time_t;

struct timespec {
        time_t  tv_sec;         /* seconds */
        long    tv_nsec;        /* nanoseconds */
};

struct super_block;

struct semaphore {
  spinlock_t lock;
  unsigned int count;
};

struct mutex {
  atomic_t count;
  spinlock_t lock;
};

//NdV: far from complete
struct device {
        struct klist            klist_children;
        struct klist_node       knode_parent;   /* node in sibling list */
        struct klist_node       knode_driver;
        struct klist_node       knode_bus;
        struct device           *parent;

        struct kobject kobj;
        char    bus_id[BUS_ID_SIZE];    /* position on parent bus */
        struct device_type      *type;
  unsigned                is_registered/*:1*/;
  unsigned                uevent_suppress/*:1*/;

        struct semaphore        sem;    /* semaphore to synchronize calls to
  //                                       * its driver.
  //                                       */

  //        struct bus_type *bus;           /* type of bus device is on */
  //      struct device_driver *driver;   /* which driver has allocated this
  //                                         device */
        void            *driver_data;   /* data private to the driver */
  //      void            *platform_data; /* Platform specific data, device
  //                                         core doesn't touch it */
  //      struct dev_pm_info      power;

  //      u64             *dma_mask;      /* dma mask (if dma'able device) */
  //      u64             coherent_dma_mask;/* Like dma_mask, but for
  //                                           alloc_coherent mappings as
  //                                           not all hardware supports
  //                                           64 bit addresses for consistent
  //                                           allocations such descriptors. */

  //        struct device_dma_parameters *dma_parms;

  //       struct list_head        dma_pools;      /* dma pools (if dma'ble) */

  //      struct dma_coherent_mem *dma_mem; /* internal for coherent mem
  //                                           override */
        /* arch specific additions */
  //      struct dev_archdata     archdata;

        spinlock_t              devres_lock;
        struct list_head        devres_head;

        /* class_device migration path */
        struct list_head        node;
        struct class            *class;
        dev_t                   devt;   /* dev_t, creates the sysfs "dev" */
  //      struct attribute_group  **groups;       /* optional groups */

        void    (*release)(struct device *dev);
};

struct device_type {
        const char *name;
        struct attribute_group **groups;
  //        int (*uevent)(struct device *dev, struct kobj_uevent_env *env);
        void (*release)(struct device *dev);
  //       int (*suspend)(struct device *dev, pm_message_t state);
        int (*resume)(struct device *dev);
};


// NdV: far from complete
struct inode {
  struct hlist_node       i_hash;
  struct list_head        i_list;
  struct list_head        i_sb_list;
  struct list_head        i_dentry;
  dev_t                   i_rdev;
  struct cdev*            i_cdev;
  unsigned long           i_ino;
  atomic_t                i_count;
  unsigned int            i_nlink;
  struct timespec         i_atime;
  struct timespec         i_mtime;
  struct super_block      *i_sb;
  struct list_head        i_devices;
  int                     i_cindex;
};

// NdV: far from complete...
struct dentry { struct inode* d_inode; };

struct vfsmount;

struct path {
  struct vfsmount *mnt;
  struct dentry *dentry;
};

typedef void (*poll_queue_proc)(struct file *, wait_queue_head_t *, struct poll_table_struct *);

typedef struct poll_table_struct {
        poll_queue_proc qproc;
} poll_table;

struct file_operations {
  struct module *owner;
  loff_t (*llseek) (struct file *, loff_t, int);
  ssize_t (*read) (struct file *, char __user *, size_t, loff_t *);
  ssize_t (*write) (struct file *, const char __user *, size_t, loff_t *);
  //ssize_t (*aio_read) (struct kiocb *, const struct iovec *, unsigned long, loff_t);
  //ssize_t (*aio_write) (struct kiocb *, const struct iovec *, unsigned long, loff_t);
  //int (*readdir) (struct file *, void *, filldir_t);
  unsigned int (*poll) (struct file *, struct poll_table_struct *);
  int (*ioctl) (struct inode *, struct file *, unsigned int, unsigned long);
  long (*unlocked_ioctl) (struct file *, unsigned int, unsigned long);
  long (*compat_ioctl) (struct file *, unsigned int, unsigned long);
  //int (*mmap) (struct file *, struct vm_area_struct *);
  int (*open) (struct inode *, struct file *);
  //int (*flush) (struct file *, fl_owner_t id);
  int (*release) (struct inode *, struct file *);
  int (*fsync) (struct file *, struct dentry *, int datasync);
  //int (*aio_fsync) (struct kiocb *, int datasync);
  int (*fasync) (int, struct file *, int);
  //int (*lock) (struct file *, int, struct file_lock *);
  //ssize_t (*sendpage) (struct file *, struct page *, int, size_t, loff_t *, int);
  //unsigned long (*get_unmapped_area)(struct file *, unsigned long, unsigned long, unsigned long, unsigned long);
  int (*check_flags)(int);
  //int (*flock) (struct file *, int, struct file_lock *);
  //ssize_t (*splice_write)(struct pipe_inode_info *, struct file *, loff_t *, size_t, unsigned int);
  //ssize_t (*splice_read)(struct file *, loff_t *, struct pipe_inode_info *, size_t, unsigned int);
  //int (*setlease)(struct file *, long, struct file_lock **);
};

extern loff_t no_llseek(struct file *file, loff_t offset, int origin);

/* Internal kernel extensions */
#define FMODE_LSEEK	4
#define FMODE_PREAD	8
#define FMODE_PWRITE	FMODE_PREAD	/* These go hand in hand */

/* File is being opened for execution. Primary users of this flag are
   distributed filesystems that can use it to achieve correct ETXTBUSY
   behavior for cross-node execution/opening_for_writing of files */
#define FMODE_EXEC	16

typedef unsigned fmode_t;

struct file {
         /*
          * fu_list becomes invalid after file_free is called and queued via
          * fu_rcuhead for RCU freeing
          */
         union {
                 struct list_head        fu_list;
                 //struct rcu_head         fu_rcuhead;
         } f_u;
  struct path             f_path;
  #define f_dentry        f_path.dentry
  #define f_vfsmnt        f_path.mnt
  const struct file_operations    *f_op;
  long          f_count;
  unsigned int            f_flags;
  fmode_t                 f_mode;
  loff_t                  f_pos;
  //         struct fown_struct      f_owner;
  //       const struct cred       *f_cred;
  //       struct file_ra_state    f_ra;

  u64                     f_version;
  //struct address_space    *f_mapping;
  void * private_data;
 };

#define fops_get(fops) \
        (((fops) && try_module_get((fops)->owner) ? (fops) : NULL))

#define TTY_LDISC_MAGIC 0x5403
#define N_TTY_BUF_SIZE 4096

#define TTY_NORMAL      0
#define TTY_BREAK	1
#define TTY_FRAME	2
#define TTY_PARITY	3
#define TTY_OVERRUN	4

struct tty_ldisc {
        int     magic;
        char    *name;
        int     num;
        int     flags;

        /*
         * The following routines are called from above.
         */
        int     (*open)(struct tty_struct *);
        void    (*close)(struct tty_struct *);
        void    (*flush_buffer)(struct tty_struct *tty);
        ssize_t (*chars_in_buffer)(struct tty_struct *tty);
        ssize_t (*read)(struct tty_struct * tty, struct file * file,
                        unsigned char __user * buf, size_t nr);
        ssize_t (*write)(struct tty_struct * tty, struct file * file,
                         const unsigned char * buf, size_t nr);
        int     (*ioctl)(struct tty_struct * tty, struct file * file,
                         unsigned int cmd, unsigned long arg);
        long    (*compat_ioctl)(struct tty_struct * tty, struct file * file,
                                unsigned int cmd, unsigned long arg);
        void    (*set_termios)(struct tty_struct *tty, struct ktermios * old);
        unsigned int (*poll)(struct tty_struct *, struct file *,
                             struct poll_table_struct *);
        int     (*hangup)(struct tty_struct *tty);

        /*
         * The following routines are called from below.
         */
        void    (*receive_buf)(struct tty_struct *, const unsigned char *cp,
                               char *fp, int count);
        void    (*write_wakeup)(struct tty_struct *);

        struct  module *owner;

        int refcount;
};

/*
struct tty_ldisc {
  struct tty_ldisc_ops* ops;
  int refcount;
};
*/

struct winsize {
  unsigned short ws_row;
  unsigned short ws_col;
  unsigned short ws_xpixel;
  unsigned short ws_ypixel;
};

struct circ_buf {
  char* buf;
  int head;
  int tail;
};

struct tasklet_struct {
  struct tasklet_struct *next;
  unsigned long state;
  atomic_t count;
  void (*func)(unsigned long);
  unsigned long data;
};

struct serial_struct {
  int     type;
  int     line;
  unsigned int    port;
  int     irq;
  int     flags;
  int     xmit_fifo_size;
  int     custom_divisor;
  int     baud_base;
  unsigned short  close_delay;
  char    io_type;
  char    reserved_char[1];
  int     hub6;
  unsigned short  closing_wait; /* time to wait before closing */
  unsigned short  closing_wait2; /* no longer used... */
  unsigned char   *iomem_base;
  unsigned short  iomem_reg_shift;
  unsigned int    port_high;
  unsigned long   iomap_base;     /* cookie passed into ioremap */
};
struct tty_operations {
  //struct tty_struct * (*lookup)(struct tty_driver *driver,
  //                              struct inode *inode, int idx);
  int  (*install)(struct tty_driver *driver, struct tty_struct *tty);
  void (*remove)(struct tty_driver *driver, struct tty_struct *tty);
  int  (*open)(struct tty_struct * tty, struct file * filp);
  void (*close)(struct tty_struct * tty, struct file * filp);
  void (*shutdown)(struct tty_struct *tty);
  int  (*write)(struct tty_struct * tty,
                const unsigned char *buf, int count);
  void  (*put_char)(struct tty_struct *tty, unsigned char ch);
  void (*flush_chars)(struct tty_struct *tty);
  int  (*write_room)(struct tty_struct *tty);
  int  (*chars_in_buffer)(struct tty_struct *tty);
  int  (*ioctl)(struct tty_struct *tty, struct file * file,
                unsigned int cmd, unsigned long arg);
  long (*compat_ioctl)(struct tty_struct *tty, struct file * file,
                       unsigned int cmd, unsigned long arg);
  void (*set_termios)(struct tty_struct *tty, struct ktermios * old);
  void (*throttle)(struct tty_struct * tty);
  void (*unthrottle)(struct tty_struct * tty);
  void (*stop)(struct tty_struct *tty);
  void (*start)(struct tty_struct *tty);
  void (*hangup)(struct tty_struct *tty);
  void (*break_ctl)(struct tty_struct *tty, int state);
  void (*flush_buffer)(struct tty_struct *tty);
  void (*set_ldisc)(struct tty_struct *tty);
  void (*wait_until_sent)(struct tty_struct *tty, int timeout);
  void (*send_xchar)(struct tty_struct *tty, char ch);
  int (*read_proc)(char *page, char **start, off_t off,
                          int count, int *eof, void *data);
  int (*write_proc)(struct file *file, const char __user *buffer,
                    unsigned long count, void *data);
  int (*tiocmget)(struct tty_struct *tty, struct file *file);
  int (*tiocmset)(struct tty_struct *tty, struct file *file,
                  unsigned int set, unsigned int clear);
  //int (*resize)(struct tty_struct *tty, struct winsize *ws);
  //int (*set_termiox)(struct tty_struct *tty, struct termiox *tnew);
#ifdef CONFIG_CONSOLE_POLL
  int (*poll_init)(struct tty_driver *driver, int line, char *options);
  int (*poll_get_char)(struct tty_driver *driver, int line);
  void (*poll_put_char)(struct tty_driver *driver, int line, char ch);
#endif
};

struct tty_driver {
  int     magic;          /* magic number for this structure */
  struct kref kref;       /* Reference management */
  struct cdev cdev;
  struct module   *owner;
  const char      *driver_name;
  const char      *name;
  int     name_base;      /* offset of printed name */
  int     major;          /* major device number */
  int     minor_start;    /* start of minor device number */
  int     minor_num;      /* number of *possible* devices */
  int     num;            /* number of devices allocated */
  short   type;           /* type of tty driver */
  short   subtype;        /* subtype of tty driver */
  struct ktermios init_termios; /* Initial termios */
  int     flags;          /* tty driver flags */
  int refcount;
  //      struct proc_dir_entry *proc_entry; /* /proc fs entry */
  struct tty_driver *other; /* only used for the PTY driver */

  /*
   * Pointer to the tty data structures
   */
  struct tty_struct **ttys;
  struct ktermios **termios;
  struct ktermios **termios_locked;
  void *driver_state;

  /*
   * Driver methods
   */
  //NdV: tty_operations replaced direct access to fcts ptrs
  // between 2.6.25 and 2.6.27.
  //const struct tty_operations *ops;
  void (*close)(struct tty_struct * tty, struct file * filp);
  void (*break_ctl)(struct tty_struct *tty, int state);
  int (*tiocmget)(struct tty_struct *tty, struct file *file);
  int (*tiocmset)(struct tty_struct *tty, struct file *file,
                  unsigned int set, unsigned int clear);
  int  (*ioctl)(struct tty_struct *tty, struct file * file,
                unsigned int cmd, unsigned long arg);
  int  (*write)(struct tty_struct * tty,
                const unsigned char *buf, int count);
  int  (*open)(struct tty_struct * tty, struct file * filp);
  void  (*put_char)(struct tty_struct *tty, unsigned char ch);
  void (*flush_chars)(struct tty_struct *tty);
  void (*set_termios)(struct tty_struct *tty, struct ktermios * old);
  int  (*write_room)(struct tty_struct *tty);
  int  (*chars_in_buffer)(struct tty_struct *tty);
  void (*throttle)(struct tty_struct * tty);
  void (*unthrottle)(struct tty_struct * tty);
  void (*stop)(struct tty_struct *tty);
  void (*start)(struct tty_struct *tty);
  void (*hangup)(struct tty_struct *tty);
  void (*flush_buffer)(struct tty_struct *tty);
  void (*set_ldisc)(struct tty_struct *tty);
  void (*wait_until_sent)(struct tty_struct *tty, int timeout);
  void (*send_xchar)(struct tty_struct *tty, char ch);
  int (*read_proc)(char *page, char **start, off_t off,
                          int count, int *eof, void *data);
  int (*write_proc)(struct file *file, const char __user *buffer,
                    unsigned long count, void *data);
  struct list_head tty_drivers;
};

typedef __u32 dev_t;

typedef unsigned int pid_t;

int prom_root_node;

/*@ assigns \nothing; */
void proc_tty_register_driver(struct tty_driver *driver);

/*@ assigns \nothing; */
void proc_tty_unregister_driver(struct tty_driver *driver);

void unregister_chrdev_region(dev_t from, unsigned count);
int cdev_add(struct cdev *p, dev_t dev, unsigned count);
void cdev_init(struct cdev *, const struct file_operations *);
int register_chrdev_region(dev_t, unsigned, const char *);

struct console {
        char    name[16];
        void    (*write)(struct console *, const char *, unsigned);
        int     (*read)(struct console *, char *, unsigned);
        struct tty_driver *(*device)(struct console *, int *);
        void    (*unblank)(void);
        int     (*setup)(struct console *, char *);
        int     (*early_setup)(void);
        short   flags;
        short   index;
        int     cflag;
        void    *data;
        struct   console *next;
};

struct tty;

struct upid { int upid_id; };

struct hlist_head {
       struct hlist_node *first;
};

struct pid
{
        atomic_t count;
        unsigned int level;
        /* lists of tasks that use this pid */
        struct hlist_head tasks[PIDTYPE_MAX];
        struct rcu_head rcu;
        struct upid numbers[1];
};

struct signal_struct {
  struct tty* tty;
   /* boolean value for session group leader */
  int leader;
  /*  struct tty_struct *tty; */ /* PH (enlevé) */
  struct pid *tty_old_pgrp;

};

typedef void __signalfn_t(int);
typedef __signalfn_t __user *__sighandler_t;

typedef void __restorefn_t(void);
typedef __restorefn_t __user *__sigrestore_t;

struct sigaction {
         __sighandler_t sa_handler;
         unsigned long sa_flags;
         __sigrestore_t sa_restorer;
         sigset_t sa_mask;               /* mask last for extensibility */
 };

 struct k_sigaction {
         struct sigaction sa;
 };


struct sighand_struct {
  atomic_t                count;
  struct k_sigaction      action[_NSIG];
  spinlock_t              siglock;
  wait_queue_head_t       signalfd_wqh;
};

struct device_attribute {
        struct attribute        attr;
  //        ssize_t (*show)(struct device *dev, struct device_attribute *attr,
  //                      char *buf);
  //      ssize_t (*store)(struct device *dev, struct device_attribute *attr,
  //                       const char *buf, size_t count);
};


struct class {
        const char              *name;
        struct module           *owner;

        struct kset             subsys;
        struct list_head        children;
        struct list_head        devices;
        struct list_head        interfaces;
        struct kset             class_dirs;
        struct semaphore        sem; /* locks children, devices, interfaces */
  //        struct class_attribute          *class_attrs;
  //      struct class_device_attribute   *class_dev_attrs;
        struct device_attribute         *dev_attrs;

  //       int (*uevent)(struct class_device *dev, struct kobj_uevent_env *env);
  //      int (*dev_uevent)(struct device *dev, struct kobj_uevent_env *env);

        void (*release)(struct class_device *dev);
        void (*class_release)(struct class *class);
        void (*dev_release)(struct device *dev);

  //    int (*suspend)(struct device *dev, pm_message_t state);
  //     int (*resume)(struct device *dev);
};

struct class_interface {
        struct list_head        node;
        struct class            *class;

  //        int (*add)      (struct class_device *, struct class_interface *);
  //      void (*remove)  (struct class_device *, struct class_interface *);
        int (*add_dev)          (struct device *, struct class_interface *);
        void (*remove_dev)      (struct device *, struct class_interface *);
};



volatile spinlock_t tasklist_lock;

unsigned long volatile jiffies;

#if 0
void *kzalloc(size_t size, unsigned flags);
void *kmalloc(size_t size, unsigned flags);
void kfree(void*);
#endif

#define kzalloc(size,flags) ({			\
  void * p = malloc(size); \
  if (p) { \
    memset(p,0,size); \
  } \
  p; \
    })

#ifdef NO_BUILTIN
char malloc_buffer[UINT_MAX];
unsigned int malloc_offset=0;
void* malloc(unsigned int size) {
  if (size==0 || UINT_MAX-malloc_offset < size) return NULL;
  unsigned int current_offset = malloc_offset;
  malloc_offset+=size;
  return &malloc_buffer[current_offset];
}
#endif

#define kmalloc(size,flags) malloc(size)
#define kfree(ptr)  (ptr ? free(ptr) : 0 )

void *get_zeroed_page(unsigned flags);


//void prefetch(void* x) { }

#define container_of(ptr, type, member) ({                      \
        const typeof( ((type *)0)->member ) *__mptr = (ptr);    \
        (type *)( (char *)__mptr - offsetof(type,member) );})

#define list_entry(ptr,type,member) container_of(ptr,type,member)
#define hlist_entry(ptr,type,member) container_of(ptr,type,member)

#define list_for_each(pos,head) \
        for (pos = (head)->next; prefetch(pos->next), pos != (head); \
                pos = pos->next)

#define list_for_each_entry(pos, head, member)                          \
        for (pos = list_entry((head)->next, typeof(*pos), member);      \
             prefetch(pos->member.next), &pos->member != (head);        \
             pos = list_entry(pos->member.next, typeof(*pos), member))

unsigned long _spin_lock_irqsave(spinlock_t *lock);
unsigned long _spin_unlock_irqrestore(spinlock_t *lock,unsigned long flags);

#define create_dir(x) 0
#define spin_lock_irqsave(lock,flags) \
  flags = _spin_lock_irqsave(lock)

#define spin_unlock_irqrestore(lock,flags) \
  _spin_unlock_irqrestore(lock,flags)

void spin_lock_init(spinlock_t* lock);
void spin_lock(spinlock_t* lock);
void spin_unlock(spinlock_t* lock);

static inline void read_lock(spinlock_t* lock) { spin_lock(lock); }
static inline void read_unlock(spinlock_t* lock) { spin_unlock(lock); }

static void init_MUTEX(struct semaphore* mutex) {
  spin_lock_init(&mutex->lock); mutex->lock = 1;
}

static inline void mutex_lock(struct mutex* mutex) {
  spin_lock(&mutex->lock);
}

static inline void mutex_unlock(struct mutex* mutex) {
  spin_unlock(&mutex->lock);
}

static inline void mutex_init(struct mutex* mutex)
{spin_lock_init(&mutex->lock);}

int prom_getint(int, char*);
int prom_getintdefault(int, char*, int);

#define BUG_ON(condition) if(condition) while(1)
#define unlikely(x) x

void panic(const char * fmt, ...);

static inline struct task_struct *next_thread(const struct task_struct *p)
{
        return list_entry(rcu_dereference(p->thread_group.next),
                          struct task_struct, thread_group);
}

extern void console_init(void);
extern void kobject_init(struct kobject *kobj, struct kobj_type *ktype);

static inline struct kset *to_kset(struct kobject *kobj)
{
        return kobj ? container_of(kobj, struct kset, kobj) : NULL;
}

struct kobject* kobject_get(struct kobject* kobj);
void kobject_put(struct kobject* kobj);

static inline struct kset *kset_get(struct kset *k)
{
        return k ? to_kset(kobject_get(&k->kobj)) : NULL;
}

static inline void kset_put(struct kset *k)
{
        kobject_put(&k->kobj);
}



struct kobj_map;
typedef struct kobject *kobj_probe_t(dev_t, int *, void *);

int kobj_map(struct kobj_map *, dev_t, unsigned long, struct module *,
	     kobj_probe_t *, int (*)(dev_t, void *), void *);

void chrdev_init(void);

#define to_dev(obj) container_of(obj, struct device, kobj)

#define POLLIN          0x0001
#define POLLPRI         0x0002
#define POLLOUT         0x0004
#define POLLERR         0x0008
#define POLLHUP         0x0010
#define POLLNVAL        0x0020

/* The rest seem to be more-or-less nonstandard. Check them! */
#define POLLRDNORM      0x0040
#define POLLRDBAND      0x0080
#ifndef POLLWRNORM
#define POLLWRNORM      0x0100
#endif
#ifndef POLLWRBAND
#define POLLWRBAND      0x0200
#endif
#ifndef POLLMSG
#define POLLMSG         0x0400
#endif
#ifndef POLLREMOVE
#define POLLREMOVE      0x1000
#endif
#ifndef POLLRDHUP
#define POLLRDHUP       0x2000
#endif

int (*platform_notify)(struct device *dev) = NULL;

void tasklet_init(struct tasklet_struct *t,
                  void (*func)(unsigned long), unsigned long data);

static inline int test_bit(unsigned int nr, unsigned long *addr)
 {
         return ((1UL << (nr % BITS_PER_LONG)) &
                 (((unsigned long *)addr)[nr / BITS_PER_LONG])) != 0;
 }

static inline void set_bit(unsigned int nr, unsigned long *addr) {
  ((unsigned long *)addr)[nr / BITS_PER_LONG] |= (1UL << (nr % BITS_PER_LONG));


}

static inline void clear_bit(unsigned int nr, unsigned long *addr) {
  ((unsigned long *)addr)[nr / BITS_PER_LONG] ^= (1UL << (nr % BITS_PER_LONG));


}

static inline int test_and_set_bit(unsigned int nr, unsigned long* addr)
{ int res = test_bit(nr,addr);
  set_bit(nr,addr);
  return res;
}

static inline int test_and_clear_bit(unsigned int nr, unsigned long* addr)
{ int res = test_bit(nr,addr);
  clear_bit(nr,addr);
  return res;
}

static inline int in_interrupt() { return 0; /* return Frama_C_nondet(0,1); */ /* otherwise may cause infinite loop */ }

// a bit short, but who cares?
#define MAX_SCHEDULE_TIMEOUT    42000

#define HZ 100

static inline long schedule_timeout(long timeout) { return 0; }

static inline int sigismember(sigset_t *set, int _sig)
{
        unsigned long sig = _sig - 1;
        if (_NSIG_WORDS == 1)
                return 1 & (set->sig[0] >> sig);
        else
                return 1 & (set->sig[sig / _NSIG_BPW] >> (sig % _NSIG_BPW));
}

extern struct file_operations def_chr_fops;

int nonseekable_open(struct inode *inode, struct file *filp);

struct device *device_create(struct class *class, struct device *parent,
                             dev_t devt, const char *fmt);

static inline void atomic_inc (atomic_t* v) { v->counter++; }

static inline struct pid *get_pid(struct pid *pid)
{
        if (pid)
                atomic_inc(&pid->count);
        return pid;
}
void put_device(struct device *dev);
void kobject_del(struct kobject *kobj);
static void device_release(struct kobject *kobj)
{
        struct device *dev = to_dev(kobj);

	/* PH (MODIF1) : code retiré */

         if (dev->release)
                 dev->release(dev);
         else if (dev->type && dev->type->release)
                 dev->type->release(dev);
         else if (dev->class && dev->class->dev_release)
                 dev->class->dev_release(dev);
         else {
                 printk(KERN_ERR "Device '%s' does not have a release() "
                         "function, it is broken and must be fixed.\n",
                         dev->bus_id);
                 WARN_ON(1);
         }
        /* PH : fin (MODIF1) */
}

static struct kobj_type device_ktype = {
        .release        = device_release,
        //        .sysfs_ops      = &dev_sysfs_ops,
};

#define to_class(obj) container_of(obj, struct class, subsys.kobj)

static void class_release(struct kobject *kobj)
{
        struct class *class = to_class(kobj);

        pr_debug("class '%s': release.\n", class->name);

        if (class->class_release)
                class->class_release(class);
        else
                pr_debug("class '%s' does not have a release() function, "
                         "be careful\n", class->name);
}


static struct kobj_type class_ktype = {
  //        .sysfs_ops      = &class_sysfs_ops,
        .release        = class_release,
};


static struct kset *class_kset;

struct class *class_create(struct module *owner, const char *name);

#define MAX_ERRNO 4095
//#define IS_ERR_VALUE(x) unlikely((x) >= (unsigned long)-MAX_ERRNO)
#define IS_ERR_VALUE(x) 0

static inline void *ERR_PTR(long error)
{
        return (void *) error;
}

static inline long PTR_ERR(const void *ptr)
{
        return (long) ptr;
}

static inline long IS_ERR(const void *ptr)
{
        return IS_ERR_VALUE((unsigned long)ptr);
}

// assume always have privileges.
static inline int capable(int cap) { return 1;}

//@ assigns \nothing;
extern struct timespec current_fs_time(struct super_block *sb);

extern size_t strlen (const char *s);
extern void *memset(void *s, int c, size_t n);
extern char *strcpy(char *restrict dest, const char *restrict src);

struct lock_class_key {
  int k;
};

int wake_up(void* a);
int file_move(struct file *file, struct list_head *list);
int spin_lock_irq(void* p);
int spin_unlock_irq(void* p);
int put_pid(void* p);
int schedule_delayed_work(void*p, int i);
int INIT_DELAYED_WORK(void* a, void* b);
void proc_tty_register_driver(struct tty_driver *driver );
int tasklet_schedule(void* p);
int remove_wait_queue(void* a,void* b);
int signal_pending(void* p);
int mutex_lock_interruptible(void* a);
int lockdep_set_class(void* a, void* b);
int request_irq(unsigned int irq, void* handler, void* a, void* devname, void* dev_id);
int n_tty_ioctl(struct tty_struct *tty , struct file *file ,
		unsigned int cmd , unsigned long arg );

int synchronize_irq(int);

#define tasklet_kill(task)
#define free_page(page)

int kernel_locked(void);

int irq_canonicalize(int);

int copy_to_user(void* dst, void* src, size_t n);

int copy_from_user(void* dst, void* src, size_t n);

int put_user(int x, int* dst);

void schedule(void);

#endif

